'''
Models is related with account management
'''

from .std import acc_name, acc_pwd, email, phone, objectid, unix_time


# database model

#:
confirm_code = {
    'type': 'string',
    'pattern': '^([a-zA-Z]{4}-){3}[a-zA-Z]{4}$'
}

# front models

#:
get_me = {
    'type': 'object',
    'required': [
        '_id', 'name', 'email', 'phone',
        'created_date', 'modified_date', 'last_action'
    ],
    'properties': {
        '_id': {'type': 'string', 'pattern': objectid},
        'name': acc_name,
        'email': email,
        'phone': phone,
        'created_date': unix_time,
        'modified_date': unix_time,
        'last_action': unix_time
    }
}

#:
post_reg_code = {
    'type': 'object',
    'required': ['name', 'email', 'pwd'],
    'properties': {
        'name': acc_name,
        'pwd': acc_pwd,
        'email': email,
        'phone': phone
    }
}

#:
post_reg = {
    'type': 'object',
    'required': ['code'],
    'properties': {
        'code': confirm_code
    }
}

#:
put_pwd = {
    'type': 'object',
    'required': ['old_pwd', 'new_pwd'],
    'properties': {
        'old_pwd': acc_pwd,
        'new_pwd': acc_pwd
    }
}

#:
post_rpwd_code = {
    'type': 'object',
    'required': ['email'],
    'properties': {
        'email': email
    }
}

#:
post_rpwd = {
    'type': 'object',
    'required': ['code', 'new_pwd'],
    'properties': {
        'code': confirm_code,
        'new_pwd': acc_pwd
    }
}



from clink.error.wsgi import WsgiResBodyError
from clink.error.http import code_to_str
from clink.iface import ILv6Handler
from clink.com import stamp


@stamp()
class SendHandler(ILv6Handler):
    '''
    Send response message to client
    '''

    def handle(self, req, res, wsgi_send):
        if res.body is not None and not isinstance(res.body, bytes):
            raise WsgiResBodyError(res.body)

        header = [(k, v) for k, v in res.header.items()]
        if res.content_type is not None:
            header.append(('Content-Type', res.content_type))
        wsgi_send(code_to_str(res.status), header)

        if res.body is None:
            return []
        else:
            return [res.body]



from clink.com import stamp, Primitive


class MongoDocSpec():
    '''
    Specify an mongo document in database
    '''

    def __init__(self, name, indexes):
        '''
        :param str name:
        :param list[pymongo.IndexModel] indexes:
        '''

        self._name = name
        self._indexes = indexes

    @property
    def name(self):
        '''
        Name of document
        '''

        return self._name

    @property
    def indexes(self):
        '''
        List of indexes of document
        '''

        return self._indexes


@stamp()
class MongoConf(Primitive):
    '''
    Specify information of mongo server

    :param str dburl:
    :param str dbname:
    '''

    def __init__(self, dburl, dbname):
        self._dburl = dburl
        self._dbname = dbname

    @property
    def dburl(self):
        '''
        URL to server, it follows forms

            - mongodb://domain/path
            - mongodb://<username>:<password>@domain/path
        '''

        return self._dburl

    @property
    def dbname(self):
        '''
        Name of database will be use
        '''

        return self._dbname



import logging
from os import path
from os.path import isfile
from stat import S_IRUSR, S_IWUSR, S_IRGRP, S_IWGRP, S_IROTH

from clink.iface import ILv8Handler
from clink.util.shell import touch
from clink.com import stamp
from clink.type import AppConf

LOGFILE_MODE = S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH


@stamp(AppConf)
class ErrorLogHandler(ILv8Handler):
    '''
    Catch error, write information to file at
    /var/tmp/<app-name>/error.log
    '''

    def __init__(self, app_conf):
        file = path.join('/var/tmp', app_conf.name, 'error.log')
        if not isfile(file):
            touch(file, LOGFILE_MODE)

        self._logger = logging.getLogger(file)
        fhandler = logging.FileHandler(file)
        formatter = logging.Formatter('%(asctime)s$ %(message)s')
        fhandler.setFormatter(formatter)
        self._logger.addHandler(fhandler)
        self._logger.setLevel(logging.INFO)

    def handle(self, req, res, e):
        msg = ' '.join([
            str(res.status), req.remote_addr, req.method, req.path,
            type(e).__name__
        ])
        self._logger.info(msg)
        return True



from smtplib import SMTP
from email.mime.text import MIMEText

from clink.com import stamp
from clink.type.com import Service
from clink.type import AppConf, AuthConf


@stamp(AppConf, AuthConf)
class SmtpSv(Service):
    '''
    Send mail message on SMTP
    '''

    def __init__(self, app_conf, auth_conf):
        '''
        :param AppConf app_conf:
        :param AuthConf auth_conf:
        '''

        self._app_conf = app_conf
        self._auth_conf = auth_conf

        self._from = '{} <{}>'.format(
            self._app_conf.name, self._auth_conf.root_email
        )

        self._smtp = self._connect()

    def send(self, dest_email, subject, text_body):
        '''
        Send plain text message

        :param str dest_email:
        :param str subject:
        :param str text_body:
        '''

        msg = MIMEText(text_body)
        msg['To'] = dest_email
        msg['Subject'] = subject
        msg['From'] = self._from

        self._smtp.sendmail(
            self._auth_conf.root_email, dest_email, msg.as_string()
        )

    def _connect(self):
        smtp = SMTP(self._auth_conf.root_email_server)
        smtp.ehlo()
        smtp.starttls()
        smtp.login(self._auth_conf.root_email, self._auth_conf.root_email_pwd)

        return smtp



from clink import stamp, mapper, Controller, AppConf


@stamp(AppConf)
@mapper.path('')
class RootCtl(Controller):
    '''
    APIs information
    '''

    def __init__(self, app_conf):
        self._info = {
            'name': app_conf.name,
            'license': app_conf.license,
            'version': str(app_conf.version),
            'organization': app_conf.org_name,
            'headquater': app_conf.org_addr
        }

    @mapper.get('')
    def api_info(self, req, res):
        res.body = self._info



#:
MIME_URLENCODE = 'application/x-www-form-urlencoded'
#:
MIME_FORMDATA = 'multipart/form-data'
#:
MIME_PLAINTEXT = 'text/plain'

MIME_EPUB = 'application/epub+zip'
#:
MIME_JAR = 'application/java-archive'
#:
MIME_JS = 'application/jsavascript'
#:
MIME_JSON = 'application/json'
#:
MIME_MSWORD = 'application/msword'
#:
MIME_BIN = 'application/octet-stream'
#:
MIME_OGG = 'application/ogg'
#:
MIME_PDF = 'application/pdf'
#:
MIME_RTF = 'application/rtf'
#:
MIME_AZW = 'application/vnd.amazon.ebook'
#:
MIME_MPKG = 'application/vnd.apple.installer+xml'
#:
MIME_XUL = 'application/vnd.moliza.xul+xml'
#:
MIME_XLS = 'application/vnd.ms-excel'
#:
MIME_PPT = 'application/vnd.ms-powerpoint'
#:
MIME_ODP = 'application/vnd.oasis.opendocument.presentaion'
#:
MIME_ODS = 'application/vnd.oasis.opendocument.spreadsheet'
#:
MIME_ODT = 'application/vnd.oasis.opendocument.text'
#:
MIME_VSD = 'application/vnd.visio'
#:
MIME_7Z = 'application/x-7z-compressed'
#:
MIME_ABW = 'application/x-abiword'
#:
MIME_BZ = 'application/x-bzip'
#:
MIME_BZ2 = 'application/x-bzip2'
#:
MIME_CSH = 'application/x-csh'
#:
MIME_RAR = 'application/x-rar-compressed'
#:
MIME_SH = 'application/x-sh'
#:
MIME_SWF = 'application/x-shockwave-flash'
#:
MIME_TAR = 'application/x-tar'
#:
MIME_XHTML = 'application/xhtml+xml'
#:
MIME_XML = 'application/xml'
#:
MIME_ZIP = 'application/zip'
#:
MIME_AAC = 'audio/acc'
#:
MIME_MIDI = 'audio/midi'
#:
MIME_OGA = 'audio/ogg'
#:
MIME_WEBA = 'audio/webm'
#:
MIME_WAV = 'audio/x-wav'
#:
MIME_TTF = 'font/ttf'
#:
MIME_WOFF = 'font/woff'
#:
MIME_WOFF2 = 'font/woff2'
#:
MIME_GIF = 'image/gif'
#:
MIME_JPEG = 'image/jpeg'
#:
MIME_SVG = 'image/svg+xml'
#:
MIME_TIFF = 'image/tiff'
#:
MIME_WEBP = 'image/webp'
#:
MIME_ICO = 'image/x-icon'
#:
MIME_ICS = 'text/calendar'
#:
MIME_CSS = 'text/css'
#:
MIME_CSV = 'text/csv'
#:
MIME_HTML = 'text/html'
#:
MIME_3GP = 'video/3gpp'
#:
MIME_3G2 = 'video/3gpp2'
#:
MIME_MPEG = 'video/mpeg'
#:
MIME_OGV = 'video/ogg'
#:
MIME_WEBM = 'video/webm'
#:
MIME_AVI = 'video/x-msvideo'



from pymongo import ASCENDING, IndexModel

from clink.service.mongo import MongoDocSpec

from clink.com import stamp
from clink.type.com import Service
from clink.service.mongo import MongoSv

ROOT_NAME = 'root'
ACC_DOCNAME = 'account'
GRP_DOCNAME = 'group'
RPWD_DOCNAME = 'rpwd'
ACCTMP_DOCNAME = 'acctmp'
DOC_NAMES = [ACC_DOCNAME, GRP_DOCNAME, RPWD_DOCNAME, ACCTMP_DOCNAME]

_ACC_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_ACC_IND_2 = IndexModel([('email', ASCENDING)], unique=True)
_ACC_DOCSPEC = MongoDocSpec(ACC_DOCNAME, [_ACC_IND_1, _ACC_IND_2])

_GRP_IND_1 = IndexModel([('name', ASCENDING)], unique=True)
_GRP_DOCSPEC = MongoDocSpec(GRP_DOCNAME, [_GRP_IND_1])

_RPWD_DOCSPEC = MongoDocSpec(RPWD_DOCNAME, [])

_ACCTMP_DOCSPEC = MongoDocSpec(ACCTMP_DOCNAME, _ACC_DOCSPEC.indexes)

_DOC_SPECS = [_ACC_DOCSPEC, _GRP_DOCSPEC, _RPWD_DOCSPEC, _ACCTMP_DOCSPEC]


@stamp(MongoSv)
class AuthDbSv(Service):
    '''
    Ensure that database is compative with both AccSv and OAuthSv
    '''

    def __init__(self, mongo_sv):
        '''
        :param MongoSv mongo_sv:
        '''

        mongo_sv.use_docspecs(_DOC_SPECS)
        self._mongo_sv = mongo_sv

    def acc_doc(self):
        '''
        Return account collection

        :rtype: pymongo.collection.Collection
        '''

        return self._mongo_sv.doc(ACC_DOCNAME)

    def grp_doc(self):
        '''
        Return group collection

        :rtype: pymongo.collection.Collection
        '''

        return self._mongo_sv.doc(GRP_DOCNAME)

    def rpwd_doc(self):
        '''
        Return refresh token collection

        :rtype: pymongo.collection.Collection
        '''

        return self._mongo_sv.doc(RPWD_DOCNAME)

    def acctmp_doc(self):
        '''
        Return temporary account collection

        :rtype: pymongo.collection.Collection
        '''

        return self._mongo_sv.doc(ACCTMP_DOCNAME)



from clink.iface import ILv5Handler
from clink.com import stamp


@stamp()
class ResCorsHandler(ILv5Handler):
    '''
    Inform client to know that server allow CORS
    '''

    def handle(self, req, res):
        if req.method.lower() != 'option':
            return
        res.header['Access-Control-Allow-Origin'] = '*'
        res.header['Access-Control-Allow-Methods'] = \
            'GET,POST,PUT,DELETE,OPTIONS'
        res.header['Access-Control-Allow-Headers'] = \
            'Authorization,Content-Type'



from clink.com import Primitive, stamp
from .version import Version


@stamp()
class AppConf(Primitive):
    '''
    Essential information about application
    '''

    def __init__(
        self, name, license='?', version=Version(0, 1, 0), 
        org_name='?', org_addr='?'
    ):
        '''
        :param str name:
        :param str license:
        :param Version version:
        :param str org_name:
        :param str org_addr:
        '''

        self.name = name
        self.license = license
        self.version = version
        self.org_name = org_name
        self.org_addr = org_addr


@stamp()
class AuthConf(Primitive):
    '''
    Authorization configuration
    '''

    def __init__(
        self, root_pwd, root_email, root_email_pwd, root_email_server,
        jwt_key, token_time=4*3600, rtoken_time=30*24*3600
    ):
        '''
        :param str root_pwd:
        :param str root_email:
        :param str root_email_pwd:
        :param str root_email_server:
        :param str jwt_key:
        :param int token_time:
        :param int rtoken_time:
        '''

        self.root_pwd = root_pwd
        self.root_email = root_email
        self.root_email_pwd = root_email_pwd
        self.root_email_server = root_email_server
        self.jwt_key = jwt_key
        self.token_time = token_time
        self.rtoken_time = rtoken_time



from string import Template
from datetime import datetime

from clink.type import AppConf, AuthConf, Service
from clink.com import stamp


@stamp(AppConf, AuthConf)
class TemplateSv(Service):
    '''
    Simple template engine
    '''

    def __init__(self, app_conf, auth_conf):
        '''
        :param AppConf app_conf:
        :param AuthConf auth_conf:
        '''

        self._base_values = {
            'app_name': app_conf.name,
            'root_name': 'root',
            'root_email': auth_conf.root_email,
            'datetime_now': datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
        }

    def build_file(self, file_path, values):
        '''
        Read data from file then build it as a template

        :param str file_path:
        :param dict values:
        :rtype: str
        '''

        f = open(file_path)
        data = f.read()
        f.close()

        return self.build_str(data, values)

    def build_str(self, data, values):
        '''
        Build string template

        :param str data:
        :param dict values:
        :rtype: str
        '''

        values.update(self._base_values)
        return Template(data).safe_substitute(values)



from .error import RouteMethodError, RoutePathError, RouteHandleError

_ALLOWED_METHODS = ['get', 'post', 'put', 'patch', 'delete', 'option', 'head']


class Route():
    '''
    Specify a route in a map
    '''

    def __init__(self, method, content_type, path, handle):
        '''
        :param str method:
        :param str content_type:
        :param function handle:
        '''

        self._verify_method(method)
        self._verify_path(path)
        self._verify_handle(handle)

        self.method = method
        self.content_type = content_type
        self.path = path
        self.handle = handle

    def _verify_path(self, path):
        if len(path) == 0:
            return
        if path[0] == '/' or path[-1:] == '/':
            raise RoutePathError(path)

    def _verify_method(self, method):
        if method.lower() not in _ALLOWED_METHODS:
            raise RouteMethodError(method)

    def _verify_handle(self, handle):
        if not callable(handle):
            raise RouteHandleError(handle)



from abc import ABC, abstractmethod
from clink.com.type import Component


class IPipeHandler(Component):
    pass


class ILv0Handler(IPipeHandler, ABC):
    '''
    Receive handling
    '''

    @abstractmethod
    def handle(self, req, res, env):
        '''
        :param Request req:
        :param Response res:
        :param dict env:
        '''

        pass


class ILv1Handler(IPipeHandler, ABC):
    '''
    Pre-Routing handling
    '''

    @abstractmethod
    def handle(self, req, res):
        '''
        :param Request req:
        :param Response res:
        '''

        pass


class ILv2Handler(IPipeHandler, ABC):
    '''
    Routing
    '''

    @abstractmethod
    def handle(self, req):
        '''
        :param Request req:
        :rtype: function
        '''

        pass


class ILv3Handler(IPipeHandler, ABC):
    '''
    Pre-Main handling
    '''

    @abstractmethod
    def handle(self, req, res):
        '''
        :param Request req:
        :param Response res:
        '''

        pass


class ILv4Handler(IPipeHandler):
    '''
    Main handling. It must be function, but we can't define interface
    for functions. Here are symbolic interface.
    '''

    pass


class ILv5Handler(IPipeHandler, ABC):
    '''
    Responding handling
    '''

    @abstractmethod
    def handle(self, req, res):
        '''
        :param Request req:
        :param Response res:
        '''

        pass


class ILv6Handler(IPipeHandler, ABC):
    '''
    Sending handling
    '''

    @abstractmethod
    def handle(self, req, res, wsgi_send):
        '''
        :param Request req:
        :param Response res:
        :param function wsgi_send:
        '''

        pass


class ILv7Handler(IPipeHandler, ABC):
    '''
    Error handling
    '''

    @abstractmethod
    def handle(self, req, res, e):
        '''
        :param Request req:
        :param Response res:
        :param Exception e:
        '''

        pass


class ILv8Handler(IPipeHandler, ABC):
    '''
    Error logging handling
    '''

    @abstractmethod
    def handle(self, req, res, e):
        '''
        :param Request req:
        :param Response res:
        :param Exception e:
        '''

        pass


class ILv9Handler(ILv6Handler):
    '''
    Sending error handling
    '''

    pass



class ExistError(Exception):
    '''
    Specify for indexes is early exist
    '''

    def __init__(self, indexes):
        '''
        :param dict indexes:
        '''

        self.indexes = indexes


class NonExistError(Exception):
    '''
    Specify for indexes doesn't exist
    '''

    def __init__(self, indexes):
        '''
        :param dict indexes:
        '''

        self.indexes = indexes


class FormatError(Exception):
    '''
    Specify for data formating is invalid
    '''

    def __init__(self, name, value, schema):
        '''
        :param str name:
        :param object value:
        :param object req:
        '''

        self.name = name
        self.value = value
        self.schema = schema

        self._msg = 'name=%s, value=%s, schema=%s' % (name, value, schema)

    def __str__(self):
        return self._msg


class ExpiredError(Exception):
    '''
    Specify expirision error
    '''

    def __init__(self, indexes):
        '''
        :param dict indexes:
        '''

        self.indexes = indexes



from .recv import RecvHandler
from .send import SendHandler

from .req_json import ReqJsonHandler
from .req_log import ReqLogHandler
from .req_url_encode import ReqUrlEncodeHandler

from .res_cors import ResCorsHandler
from .res_json import ResJsonHandler

from .err_log import ErrorLogHandler
from .err_http import ErrorHttpHandler
from .dflow_err import DflowErrorHandler
from .routing_err import RoutingErrorHandler



from .type import Component, COM_ATTR, COM_DEP


class PrimError(Exception):
    def __init__(self, prim_type, msg):
        self.prim_type = prim_type
        self._msg = '%s %s' % (prim_type, msg)

    def __str__(self):
        return self._msg


class ComDepedencyError(Exception):
    def __init__(self, com_type, msg):
        self._msg = '%s: %s' % (com_type.__name__, msg)

    def __str__(self):
        return self._msg


class InjectorLoadingError(Exception):
    pass


class ComTypeError(Exception):
    def __init__(self, com_type):
        self._msg = '%s must extend from %s' % (type(Component))

    def __str__(self):
        return self._msg


class ComAttrError(Exception):
    def __init__(self, com_type):
        self._msg = '%s must contains %s.%s' (COM_ATTR, COM_DEP)

    def __str__(self):
        return self._msg


class CircleComError(Exception):
    def __init__(self, com_types):
        self._msg = ', '.join([t.__name__ for t in com_types])

    def __str__(self):
        return self._msg


class ComExistError(Exception):
    def __init__(self, com_type):
        self._msg = com_type.__name__

    def __str__(self):
        return self._msg


class ComCreationError(Exception):
    def __init__(self, com_type, args):
        self._msg = 'type=%s, args=%s' % (com_type, args)

    def __str__(self):
        return self._msg



from clink.error.http import Http404Error, Http400Error, Http403Error
from clink import stamp, mapper, AppConf, AuthConf, Controller
from clink.service import AccSv, TemplateSv, SmtpSv, OAuthSv
from clink.util import fs


@stamp(AppConf, AuthConf, AccSv, OAuthSv, SmtpSv, TemplateSv)
@mapper.path('acc')
class AccCtl(Controller):
    '''
    Manage accounts and related concepts
    '''

    def __init__(
        self, app_conf, auth_conf, acc_sv, oauth_sv, smtp_sv, tpl_sv
    ):
        self._app_conf = app_conf
        self._auth_conf = auth_conf
        self._acc_sv = acc_sv
        self._oauth_sv = oauth_sv
        self._smtp_sv = smtp_sv
        self._tpl_sv = tpl_sv

        self._init_tpl_files()

    @mapper.post('reg/code')
    def mk_reg_code(self, req, res):
        if req.body is None:
            raise Http400Error(req, 'Require name, pwd, email')
        if 'name' not in req.body:
            raise Http400Error(req, 'Require name')
        if 'pwd' not in req.body:
            raise Http400Error(req, 'Require pwd')
        if 'email' not in req.body:
            raise Http400Error(req, 'Require email')

        result = self._acc_sv.mk_reg_code(
            req.body['name'], req.body['pwd'], req.body['email']
        )

        values = {
            'reg_code': result.code,
            'acc_name': req.body['name'],
            'acc_email': req.body['email'],
            'expired_date': result.exp_date.strftime('%Y-%m-%d %H:%M:%S'),
            'remote_addr': req.remote_addr
        }
        txt_body = self._tpl_sv.build_file(self._REG_CODE_TPL, values)
        subject = 'Registration'
        self._smtp_sv.send(req.body['email'], subject, txt_body)

        res.status = 204

    @mapper.post('reg')
    def confirm_reg_code(self, req, res):
        if req.body is None or 'code' not in req.body:
            raise Http400Error(req, 'Require code')
        reg_code = req.body['code']

        acc = self._acc_sv.cf_reg_code(reg_code)

        values = {
            'acc_name': acc['name'],
            'acc_email': acc['email'],
            'remote_addr': req.remote_addr
        }
        txt_body = self._tpl_sv.build_file(self._REG_TPL, values)
        subject = 'Registration'
        self._smtp_sv.send(acc['email'], subject, txt_body)

        res.status = 204

    @mapper.get('me')
    def acc_info(self, req, res):
        acc_id = self._oauth_sv.authen_req(req)
        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http404Error(req, 'Identity %s invalid' % str(acc_id))

        res.body = {
            '_id': str(acc['_id']),
            'name': acc['name'],
            'email': acc['email'],
            'phone': acc['phone'],
            'created_date': int(acc['created_date'].timestamp()),
            'modifired_date': int(acc['modified_date'].timestamp()),
            'last_action': acc['last_action']
        }

    @mapper.put('me/pwd')
    def change_pwd(self, req, res):
        if req.body is None:
            raise Http400Error(req, 'Require old_pwd, new_pwd')
        if 'old_pwd' not in req.body:
            raise Http400Error(req, 'Require old_pwd')
        if 'new_pwd' not in req.body:
            raise Http400Error(req, 'Require new_pwd')
        old_pwd = req.body['old_pwd']
        new_pwd = req.body['new_pwd']

        acc_id = self._oauth_sv.authen_req(req)
        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http404Error(req, 'Not found identity %s' % str(acc_id))
        acc = self._acc_sv.find_pwd(acc['name'], old_pwd)
        if acc is None:
            raise Http403Error(req, 'Old password is invalid')

        self._acc_sv.ch_pwd(acc_id, new_pwd)

        values = {
            'acc_name': acc['name'],
            'remote_addr': req.remote_addr
        }
        txt_body = self._tpl_sv.build_file(self._CHPWD_TPL, values)
        subject = 'Change password'
        self._smtp_sv.send(acc['email'], subject, txt_body)

        res.status = 204

    @mapper.post('pwd/code')
    def mk_reset_pwd_code(self, req, res):
        if req.body is None or 'email' not in req.body:
            raise Http400Error(req, 'Require email')
        email = req.body['email']

        acc = self._acc_sv.find_email(email)
        if acc is None:
            raise Http404Error(req, 'Email does not exist')

        result = self._acc_sv.mk_rpwd_code(email)

        values = {
            'reset_pwd_code': result.code,
            'app_name': self._app_conf.name,
            'acc_name': acc['name'],
            'expired_date': result.exp_date.strftime('%Y-%m-%d %H:%M:%S'),
            'remote_addr': req.remote_addr
        }

        txt_body = self._tpl_sv.build_file(self._RPWD_CODE_TPL, values)
        subject = 'Reset password code'
        self._smtp_sv.send(email, subject, txt_body)

        res.status = 204

    @mapper.post('pwd')
    def confirm_reset_pwd_code(self, req, res):
        if req.body is None:
            raise Http400Error(req, 'Require code, new_pwd')
        if 'code' not in req.body:
            raise Http400Error(req, 'Require code')
        if 'new_pwd' not in req.body:
            raise Http400Error(req, 'Require new_pwd')
        reset_code = req.body['code']
        new_pwd = req.body['new_pwd']

        acc_id = self._acc_sv.cf_rpwd_code(reset_code, new_pwd)

        acc = self._acc_sv.find_id(acc_id)
        if acc is None:
            raise Http404Error(req, 'Not found identity %s' % str(acc_id))

        values = {
            'new_pwd': new_pwd,
            'reset_pwd_code': reset_code,
            'acc_name': acc['name'],
            'remote_addr': req.remote_addr
        }

        txt_msg = self._tpl_sv.build_file(self._RPWD_TPL, values)
        subject = 'Reset password'
        self._smtp_sv.send(acc['email'], subject, txt_msg)

        res.status = 204

    def _init_tpl_files(self):
        '''
        Detect template files
        '''

        self._REG_CODE_TPL = fs.asset_path('tpl/reg-code.txt')
        self._REG_TPL = fs.asset_path('tpl/reg-tmp.txt')
        self._CHPWD_TPL = fs.asset_path('tpl/change-pwd.txt')
        self._RPWD_TPL = fs.asset_path('tpl/reset-pwd.txt')
        self._RPWD_CODE_TPL = fs.asset_path('tpl/reset-pwd-code.txt')



import random
from datetime import datetime, timedelta
from time import time
from string import ascii_lowercase, ascii_uppercase, digits
from hashlib import sha224

from clink.type import Service, AuthConf
from clink.com import stamp
from clink.dflow import verify, ExistError, NonExistError, ExpiredError
from clink.model.std import acc_name as name_model, acc_pwd as pwd_model
from clink.model.std import email as email_model, phone as phone_model
from clink.model.acc import confirm_code as confirm_code_model

from .authdb_sv import AuthDbSv
from .type import ConfirmCodeSpec

_ACT_REGISTERED = 'REGISTERED'
_ACT_CHANGE_PWD = 'CHANGE_PWD'
_ACT_RESET_PWD = 'RESET_PWD'
_ACT_ADD_TO_GRP = 'ADD_TO_GRP'
_ACT_RM_FRM_GRP = 'RM_FRM_GRP'

_PWD_CHARS = ascii_lowercase + ascii_uppercase + digits
_CODE_CHARS = ascii_uppercase


def _hash_pwd(password):
    return sha224(password.encode('utf-8')).hexdigest()


def _rand_pwd():
    return ''.join(random.sample(_PWD_CHARS, 6))


def rand_code():
    a = ''.join(random.sample(_CODE_CHARS, 4))
    b = ''.join(random.sample(_CODE_CHARS, 4))
    c = ''.join(random.sample(_CODE_CHARS, 4))
    d = ''.join(random.sample(_CODE_CHARS, 4))

    return '-'.join([a, b, c, d])


@stamp(AuthDbSv, AuthConf)
class AccSv(Service):
    '''
    Manage accounts and related concepts
    '''

    def __init__(self, authdb_sv, auth_conf):
        '''
        :param AuthDbSv authdb_sv:
        :param AuthConf auth_conf:
        '''

        self._acc_doc = authdb_sv.acc_doc()
        self._grp_doc = authdb_sv.grp_doc()
        self._rpwd_doc = authdb_sv.rpwd_doc()
        self._acctmp_doc = authdb_sv.acctmp_doc()

        self.rpwd_time = 3600
        self.create_time = 3600

        root_acc = self.find_name('root')
        if root_acc is None:
            self.mk_acc('root', auth_conf.root_pwd, auth_conf.root_email)

    @verify(None, name_model, pwd_model, email_model, phone_model)
    def mk_acc(self, name, password, email, phone=None):
        '''
        Create new account

        :param str name:
        :param str password:
        :param str email:
        :param str phone:
        :rtype: bson.objectid.ObjectId
        :raise TypeError:
        '''

        account = {
            'name': name,
            'hashpwd': _hash_pwd(password),
            'email': email,
            'phone': phone,
            'groups': [],
            'created_date': datetime.utcnow(),
            'modified_date': datetime.utcnow(),
            'last_action': _ACT_REGISTERED
        }
        result = self._acc_doc.insert_one(account)
        return result.inserted_id

    @verify(None, name_model, pwd_model, email_model, phone_model)
    def mk_reg_code(self, name, password, email, phone=None):
        '''
        Create a registration code. Use returned code with cf_reg_code()
        to create account

        :param str name:
        :param str password:
        :param str email:
        :param str phone:
        :rtype: ConfirmCodeSpec
        :raise TypeError:
        :raise ExistError:
        '''

        if self._acc_doc.find_one({'name': name}) is not None:
            raise ExistError({'name': name})
        if self._acc_doc.find_one({'email': email}) is not None:
            raise ExistError({'email': email})
        if phone is not None:
            if self._acc_doc.find_one({'phone': phone}) is not None:
                raise ExistError({'phone': phone})

        if self._acctmp_doc.find_one({'name': name}) is not None:
            raise ExistError({'name': name})
        if self._acctmp_doc.find_one({'email': email}) is not None:
            raise ExistError({'email': email})
        if phone is not None:
            if self._acctmp_doc.find_one({'phone': phone}) is not None:
                raise ExistError({'phone': 'phone'})

        datetime_now = datetime.utcnow().timestamp()
        self._acctmp_doc.delete_many({'_expired_date': {'$lt': datetime_now}})

        creation_code = rand_code()
        expired_date = datetime.utcnow() + timedelta(hours=self.create_time)
        acctmp = {
            'name': name,
            'hashpwd': _hash_pwd(password),
            'email': email,
            'phone': phone,
            'groups': [],
            'created_date': datetime.utcnow(),
            'modified_date': datetime.utcnow(),
            'last_action': _ACT_REGISTERED,

            '_expired_date': expired_date.timestamp(),
            '_creation_code': creation_code
        }
        self._acctmp_doc.insert_one(acctmp)

        return ConfirmCodeSpec(creation_code, expired_date)

    @verify(None, confirm_code_model)
    def cf_reg_code(self, code):
        '''
        Use registration code to create account

        :param str code:
        :rtype: dict
        :raise ExistError:
        :raise ExpiredError:
        '''

        acctmp = self._acctmp_doc.find_one({'_creation_code': code})
        if acctmp is None:
            raise NonExistError({'code': code})
        if acctmp['_expired_date'] < datetime.utcnow().timestamp():
            raise ExpiredError({'code': time()})
        self._acctmp_doc.delete_one({'_creation_code': code})

        del acctmp['_id']
        del acctmp['_expired_date']
        del acctmp['_creation_code']

        self._acc_doc.insert_one(acctmp)
        del acctmp['hashpwd']

        return acctmp

    def find_id(self, id):
        '''
        Find account by identity

        :param bson.objectid.ObjectId id:
        :rtype: dict
        '''

        return self._acc_doc.find_one({'_id': id})

    @verify(None, name_model)
    def find_name(self, name):
        '''
        Find account by name

        :param str name:
        :rtype: dict
        :raise TypeError:
        '''

        return self._acc_doc.find_one({'name': name})

    @verify(None, email_model)
    def find_email(self, email):
        '''
        Find account by email

        :param str email:
        :rtype: dict
        :raise TypeError:
        '''

        return self._acc_doc.find_one({'email': email})

    @verify(None, phone_model)
    def find_phone(self, phone):
        '''
        Find account by phone number

        :param str phone:
        :rtype: dict
        :raise TypeError:
        '''

        return self._acc_doc.find_one({'phone': phone})

    @verify(None, name_model, pwd_model)
    def find_pwd(self, name, pwd):
        '''
        Find account by name and password

        :param str name:
        :param str pwd:
        :rtype: dict
        :raise TypeError:
        '''

        hashpwd = _hash_pwd(pwd)
        return self._acc_doc.find_one({'name': name, 'hashpwd': hashpwd})

    def rm_acc(self, id):
        '''
        Remove account by identity

        :param bson.objectid.ObjectId id:
        '''

        result = self._acc_doc.delete_one({'_id': id})
        if result.deleted_count != 1:
            raise NonExistError({'id': id})

    @verify(None, None, pwd_model)
    def ch_pwd(self, id, new_pwd):
        '''
        Change password of account by identity

        :param bson.objectid.ObjectId id:
        :param str new_pwd:
        :raise TypeError:
        '''

        upd = {
            '$set': {
                'hashpwd': _hash_pwd(new_pwd),
                'modified_date': datetime.utcnow(),
                'last_action': _ACT_CHANGE_PWD
            }
        }
        result = self._acc_doc.update_one({'_id': id}, upd)

        if result.modified_count != 1:
            raise NonExistError({'id': id})

    @verify(None, email_model)
    def mk_rpwd_code(self, email):
        '''
        Create reset password code from email.
        Use returned code with cf_rpwd_code() to reset to new password

        :param str email:
        :rtype: ConfirmCodeSpec
        :raise TypeError:
        '''

        acc = self._acc_doc.find_one({'email': email})
        if acc is None:
            raise NonExistError({'email': email})
        self._rpwd_doc.delete_many({'acc_id': acc['_id']})

        reset_code = rand_code()
        exp_date = datetime.utcnow() + timedelta(hours=self.rpwd_time)
        code_spec = {
            'code': reset_code,
            'expired_date': exp_date.timestamp(),
            'acc_id': acc['_id'],
            'acc_email': acc['email']
        }
        self._rpwd_doc.insert_one(code_spec)

        return ConfirmCodeSpec(reset_code, exp_date)

    @verify(None, confirm_code_model, pwd_model)
    def cf_rpwd_code(self, code, new_pwd):
        '''
        Reset password from code

        :param str code:
        :param str new_pwd:
        :rtype: bson.objectid.ObjectId
        :raise TypeError:
        '''

        code_spec = self._rpwd_doc.find_one({'code': code})
        if code_spec is None:
            raise NonExistError({'code': code})
        if code_spec['expired_date'] < time():
            raise ExpiredError({'code': time()})

        acc_id = code_spec['acc_id']
        self._rpwd_doc.delete_many({'acc_id': acc_id})

        new_hashpwd = _hash_pwd(new_pwd)
        upd = {
            '$set': {
                'hashpwd': new_hashpwd,
                'last_action': _ACT_RESET_PWD
            }
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)

        return acc_id

    def mk_group(self, group_name):
        '''
        Create new account group

        :param str group_name:
        '''

        self._grp_doc.insert_one({'name': group_name})

    def rm_group(self, group_name):
        '''
        Remove account group

        :param str group_name:
        '''

        result = self._grp_doc.delete_one({'name': group_name})

        if result.deleted_count != 1:
            raise NonExistError({'group_name': group_name})

    def add_to_group(self, acc_id, group_name):
        '''
        Put an account into group

        :param bson.objectid.ObjectId acc_id:
        :param str group_name:
        '''

        acc = self._acc_doc.find_one({'_id': acc_id})
        if acc is None:
            raise NonExistError({'id': acc_id})

        grp = self._grp_doc.find_one({'name': group_name})
        if grp is None:
            raise NonExistError({'group_name': group_name})
        if group_name in acc['groups']:
            raise ExistError({'group_name': group_name})

        upd = {
            '$push': {'groups': group_name},
            '$set': {'last_action': _ACT_ADD_TO_GRP}
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)

    def del_fm_group(self, acc_id, group_name):
        '''
        Remove an account from group

        :param bson.objectid.ObjectId acc_id:
        :param str group_name:
        '''

        acc = self._acc_doc.find_one({'_id': acc_id})

        if acc is None:
            raise NonExistError({'id': acc_id})
        if group_name not in acc['groups']:
            raise NonExistError({'group_name': group_name})

        upd = {
            '$pull': {'groups': group_name},
            '$set': {'last_action': _ACT_RM_FRM_GRP}
        }
        self._acc_doc.update_one({'_id': acc_id}, upd)



from pymongo import MongoClient, IndexModel
from clink.com import stamp, Component

from .error import DocumentNotExist, DocumentIndexError
from .type import MongoConf


@stamp(MongoConf)
class MongoSv(Component):
    '''
    MongoDB interface in application layer
    '''

    def __init__(self, conf):
        '''
        :param MongoConf conf:
        '''

        self._conf = conf
        self._doc_specs = []
        self._client = None
        self._doc_names = []

    def use_docspec(self, doc_spec):
        '''
        Put a document's specification under management

        :param MongoDocSpec doc_spec:
        '''

        db = self._instance()
        db_docs = db.collection_names()

        if doc_spec.name in db_docs:
            self._verify_spec(doc_spec)
        else:
            self._create_spec(doc_spec)

        self._doc_names.append(doc_spec.name)

    def use_docspecs(self, doc_specs):
        '''
        Put document's specifications under management

        :param list[MongoDocSpec] doc_specs:
        '''

        for spec in doc_specs:
            self.use_docspec(spec)

    def doc(self, name):
        '''
        Return document object

        :param str name:
        :rtype: pymongo.collection.Collection
        :raise DocumentNotExist:
        '''

        if name not in self._doc_names:
            raise DocumentNotExist(name)
        return self._instance()[name]

    def docs(self, *args):
        '''
        Return document objects

        :param tuple[str] args:
        :rtype: tuple[pymongo.collection.Collection]
        :raise DocumentNotExist:
        '''
        for name in args:
            if name not in self._doc_names:
                raise DocumentNotExist(name)
        db = self._instance()
        return tuple([db[name] for name in args])

    def close(self):
        '''
        Close connection to server
        '''

        if self._client is not None:
            self._client.close()

    def clear(self):
        '''
        Clear all of data in database
        '''

        self._connect()
        self._client.drop_database(self._conf.dbname)

    def _create_spec(self, doc_spec):
        db = self._instance()
        doc = db[doc_spec.name]

        if len(doc_spec.indexes) > 0:
            doc.create_indexes(doc_spec.indexes)

    def _verify_spec(self, doc_spec):
        '''
        Verify that document's specification is valid in database

        :param MongoDocSpec:
        :raise DocumentIndexError:
        '''

        db = self._instance()
        doc_index_info = db[doc_spec.name].index_information()
        for index in doc_spec.indexes:
            # check index is exist
            if index.document['name'] not in doc_index_info:
                raise DocumentIndexError(doc_spec.name, index)

            # reconstruct index model from information
            index_name = index.document['name']
            doc_info = doc_index_info[index_name]
            doc_info_key = doc_info['key']
            doc_info['name'] = index_name
            del doc_info['key']
            doc_index = IndexModel(doc_info_key, **doc_info)

            # compare two index
            if not self._index_is_equal(index, doc_index):
                raise DocumentIndexError(doc_spec.name, index)

    def _index_is_equal(self, index_1, index_2):
        '''
        Compare between two indexes

        :param pymongo.IndexModel index_1:
        :param pymongo.IndexModel index_2:
        :rtype: bool
        '''

        # index_1, index_2: pymongo.IndexModel
        doc_1 = index_1.document
        doc_2 = index_2.document
        for k in doc_1:
            if k not in doc_2:
                return False
            if doc_1[k] != doc_2[k]:
                return False
        return True

    def _connect(self):
        '''
        Make sure that client connect to server
        '''

        if self._client is None:
            self._client = MongoClient(host=self._conf.dburl)

    def _instance(self):
        '''
        Return instance of database object

        :rtype: pymongo.database.Database
        '''

        self._connect()
        return self._client[self._conf.dbname]



from clink.mime.type import MIME_JSON
from clink.com.label import write_stamp
from .type import CtlMethod

CTL_PATH_ATTR = 'ctl_path'
CTL_METHOD_ATTR = 'ctl_method'


def path(path):
    '''
    Specify a path to controller

    :param str path:
    '''

    def decorator_fn(target):
        write_stamp(target, CTL_PATH_ATTR, path)
        return target
    return decorator_fn


def map(method, path, content_type):
    '''
    Specify a controller's method

    :param str method:
    :param str path:
    :param str content_type:
    '''

    def decorator_fn(target):
        ctl_method = CtlMethod(method, path, content_type)
        write_stamp(target, CTL_METHOD_ATTR, ctl_method)
        return target
    return decorator_fn


def get(path):
    '''
    Specify a GET controller's method

    :param str path:
    '''

    return map('get', path, None)


def post(path, content_type=MIME_JSON):
    '''
    Map a POST controller's method

    :param str path:
    :param str content_type:
    '''

    return map('post', path, content_type)


def put(path, content_type=MIME_JSON):
    '''
    Map a PUT controller's method

    :param str path:
    :param str content_type:
    '''

    return map('put', path, content_type)


def patch(path, content_type=MIME_JSON):
    '''
    Map a PATCH controller's method

    :param str path:
    :param str content_type:
    '''

    return map('patch', path, content_type)


def delete(path):
    '''
    Map a DELETE controller's method

    :param str path:
    :param str content_type:
    '''

    return map('delete', path, None)



import json

from clink.error.http import code_to_str
from clink.iface import ILv7Handler
from clink.mime.type import MIME_JSON
from clink.com import stamp
from clink.routing.error import PathNotFoundError, HandleNotFoundError


@stamp()
class RoutingErrorHandler(ILv7Handler):
    '''
    Catch routing error and send information to client
    '''

    def handle(self, req, res, e):
        if isinstance(e, PathNotFoundError):
            res.status = 404
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 404,
                'status_name': code_to_str(404),
                'message': 'Path does not exist'
            }).encode('utf-8')
            return True
        elif isinstance(e, HandleNotFoundError):
            res.status = 405
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 405,
                'status_name': code_to_str(405),
                'message': 'No handler for method'
            }).encode('utf-8')
            return True

        return False



class Version():
    '''
    Strict version
    '''

    def __init__(self, major, minor, rev):
        '''
        :param int major:
        :param int minor:
        :param int rev:
        '''

        self.major = major
        self.minor = minor
        self.rev = rev

    def __str__(self):
        return '%s.%s.%s' % (self.major, self.minor, self.rev)



import json

from clink.error.http import Http400Error
from clink.iface import ILv3Handler
from clink.mime.type import MIME_JSON
from clink.com import stamp


@stamp()
class ReqJsonHandler(ILv3Handler):
    '''
    Map JSON string from body message to Python object
    '''

    def handle(self, req, res):
        if req.content_type != MIME_JSON:
            return
        if req.body is None:
            return
        try:
            req.body = json.loads(req.body.decode('utf-8'))
        except ValueError:
            raise Http400Error(req, 'body is invalid json format')



import json

from clink.error.http import HttpError, code_to_str
from clink.iface import ILv7Handler
from clink.mime.type import MIME_JSON
from clink.com import stamp


@stamp()
class ErrorHttpHandler(ILv7Handler):
    '''
    Catch HTTP error and make response message correspond with
    error
    '''

    def handle(self, req, res, e):
        if not isinstance(e, HttpError):
            return False
        res.status = e.status
        res.header = {}
        res.content_type = MIME_JSON
        res.body = json.dumps({
            'status': e.status,
            'status_name': code_to_str(e.status),
            'message': e.msg
        }).encode('utf-8')
        return True



from os import path
from os.path import realpath, dirname

_ASSET_DIR = realpath(path.join(dirname(__file__), '../asset'))


def asset_path(rel_path):
    '''
    Return absolute path to Clink's asset directory

    :param str path:
    :rtype: str
    '''

    return path.join(_ASSET_DIR, rel_path)



from functools import wraps
from inspect import getargspec
from jsonschema import ValidationError, validate

from .error import FormatError


def verify(*schemas):
    '''
    Decorator, verify formating of input arguments

    :param tuple[dict] schemas:
    :rtype: function
    '''

    def decorator(target):
        argspec = getargspec(target)

        @wraps(target)
        def new_fn(*args, **kargs):
            fargs = args
            oargs_len = len(argspec.args)
            args_len = len(args)

            if args_len < oargs_len:
                fargs = list(args)
                apd_len = oargs_len - args_len - 1
                apd_args = argspec.defaults[-apd_len:]
                fargs.extend(apd_args)
                fargs = tuple(fargs)
            if len(schemas) != len(fargs):
                raise IndexError(schemas, fargs)

            for arg_index, (schema, arg) in enumerate(zip(schemas, fargs)):
                if schema is None:
                    continue
                try:
                    validate(arg, schema)
                except ValidationError as e:
                    name = '.'.join(e.absolute_path)
                    if len(name) == 0:
                        name = argspec.args[arg_index]
                    raise FormatError(name, e.instance, e.schema)
            return target(*args, **kargs)
        return new_fn
    return decorator



import jwt
from time import time
from bson import ObjectId
from jwt.exceptions import ExpiredSignatureError, DecodeError

from clink.type import AuthConf
from clink.com import stamp
from clink.type.com import Service
from clink.error.http import Http400Error, Http401Error
from clink.dflow import verify, NonExistError, ExpiredError, FormatError

from .authdb_sv import AuthDbSv
from .acc_sv import AccSv
from clink.model.std import acc_name as acc_name_model, acc_pwd as pwd_model


@stamp(AuthDbSv, AccSv, AuthConf)
class OAuthSv(Service):
    '''
    Limited OAuth2 implementation
    '''

    _TOKEN_ALG = 'HS512'

    def __init__(self, authdb_sv, acc_sv, auth_conf):
        '''
        :param AuthDbSv authdb_sv:
        :param AccSv acc_sv:
        :param AuthConf auth_conf:
        '''

        self._authdb_sv = authdb_sv
        self._acc_sv = acc_sv

        self._jwt_key = auth_conf.jwt_key
        self._token_time = auth_conf.token_time
        self._rtoken_time = auth_conf.rtoken_time

    @verify(None, acc_name_model, pwd_model)
    def mktoken_pwd(self, name, password):
        '''
        Create an token from account name and password

        :param str name:
        :param str password:
        :rtype: dict
        :raise NonExistError:
        '''

        acc = self._acc_sv.find_pwd(name, password)
        if acc is None:
            raise NonExistError({'name': name, 'password': '***'})

        return self._mk_token(acc['_id'])

    def mktoken_rtoken(self, rtoken):
        '''
        Create an token from refresh token

        :param str rtoken:
        :rtype: dict
        :raise TypeError:
        :raise ExpiredError:
        '''

        try:
            rtoken_raw = jwt.decode(
                rtoken, self._jwt_key, algorithm=self._TOKEN_ALG
            )
            return self._mk_token(rtoken_raw['sub'])
        except ExpiredSignatureError:
            raise ExpiredError({'refresh_token': None})
        except DecodeError:
            raise FormatError('refresh_token', None, None)

    def authen(self, access_token):
        '''
        Authenticate access token

        :param str access_token:
        :rtype: bson.objectid.ObjectId
        :raise FormatError:
        :raise ExpiredError:
        '''

        try:
            atoken_raw = jwt.decode(
                access_token, self._jwt_key, algorithm=self._TOKEN_ALG
            )
            return ObjectId(atoken_raw['sub'])
        except ExpiredSignatureError:
            raise ExpiredError({'access_token': None})
        except DecodeError:
            raise FormatError('access_token', None, None)

    def authen_req(self, req):
        '''
        Authenticate HTTP request

        :param Request req:
        :rtype mongo.objectid.ObjectId:
        :raise Http400Error:
        '''

        if 'AUTHORIZATION' not in req.header:
            raise Http401Error(req)
        auth_header = req.header['AUTHORIZATION']
        auth_type = auth_header[:7]
        if auth_type != 'Bearer ':
            raise Http400Error(req)
        atoken = auth_header[7:]

        return self.authen(atoken)

    def _mk_token(self, acc_id):
        '''
        Create token

        :param bson.objectid.ObjectId:
        :rtype: dict
        '''

        return {
            'token_type': 'Bearer',
            'expires_in': time() + self._token_time,
            'access_token': self._mk_atoken(acc_id),
            'refresh_token': self._mk_rtoken(acc_id)
        }

    def _mk_atoken(self, acc_id):
        '''
        Create access token

        :param bson.object.ObjectId:
        :rtype: str
        '''

        token_raw = {
            'sub': str(acc_id),
            'exp': time() + self._token_time
        }
        token = jwt.encode(
            token_raw, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        return token.decode()

    def _mk_rtoken(self, acc_id):
        '''
        Create refresh token

        :param bson.object.ObjectId:
        :rtype: str
        '''

        token_raw = {
            'sub': str(acc_id),
            'exp': time() + self._rtoken_time
        }
        token = jwt.encode(
            token_raw, self._jwt_key, algorithm=self._TOKEN_ALG
        )
        return token.decode()



class DocSpecExit(Exception):
    def __init__(self, name):
        self._msg = name

    def __str__(self):
        return self._msg


class DocumentNotExist(Exception):
    def __init__(self, doc_name):
        self._doc_name = doc_name

    def __str__(self):
        return self._doc_name


class DocumentIndexError(Exception):
    def __init__(self, doc_name, req_index):
        index_doc = req_index.document
        attr_unique = None
        if 'unique' in index_doc:
            attr_unique = index_doc['unique']
        attr_min = None
        if 'min' in index_doc:
            attr_min = index_doc['min']
        attr_max = None
        if 'max' in index_doc:
            attr_max = index_doc['max']
        index_str = 'name={}; unique={}; min={}; max={}; key={};'.format(
            index_doc['name'], attr_unique, attr_min, attr_max,
            index_doc['key']
        )
        self._msg = 'document \'{}\' must be specify indexes: {}'.format(
            doc_name, index_str
        )

    def __str__(self):
        return self._msg



class CtlSpecError(Exception):
    def __init__(self, ctl):
        self._msg = '%s MUST contains COM_ATT.ROUTE_PATH_ATTR' % ctl.__name__

    def __str__(self):
        return self._msg


class PathNotFoundError(Exception):
    def __init__(self, path):
        self._msg = path

    def __str__(self):
        return self._msg


class HandleNotFoundError(Exception):
    def __init__(self, method, content_type, path):
        self._msg = '%s:%s:%s' % (method, content_type, path)

    def __str__(self):
        return self._msg


class RouteExistError(Exception):
    def __init__(self, route):
        self._msg = '%s %s %s' % (
            route.method, route.content_type, route.path
        )

    def __str__(self):
        return self._msg


class RouteMethodError(Exception):
    def __init__(self, method):
        self._msg = 'Method %s not allowed' % method

    def __str__(self):
        return self._msg


class RoutePathError(Exception):
    def __init__(self, path):
        self._msg = '%s must not contains slash at begin and end' % path

    def __str__(self):
        return self._msg


class RouteHandleError(Exception):
    def __init__(self, handle):
        self._msg = 'Handle %s must be callable' % handle

    def __str__(self):
        return self._msg



from bson import json_util

from clink.iface import ILv5Handler
from clink.mime.type import MIME_JSON
from clink.com import stamp


@stamp()
class ResJsonHandler(ILv5Handler):
    '''
    Serialize Python object to JSON string
    '''

    def handle(self, req, res):
        if res.content_type != MIME_JSON:
            return
        if res.body is None:
            return
        res.body = json_util.dumps(res.body).encode('utf-8')



from collections import deque

from .label import read_stamp, COM_DPDC_ATTR
from .type import Primitive, Component
from .error import CircleComError, ComExistError, ComDepedencyError, \
                   ComCreationError, InjectorLoadingError, PrimError


class Injector():
    '''
    Object management
    '''

    def __init__(self):
        self._com_dict = {}
        self._com_layer = []
        self.com_inst = {}
        self._loaded = False

    def add_prim(self, com_obj):
        '''
        Put an instance of component under management. Instance MUST be
        primitive

        :param clink.com.Component com_obj:
        :raise ComExistError:
        :raise ComDepedencyError:
        '''

        com_type = type(com_obj)
        if not issubclass(com_type, Primitive):
            raise TypeError(
                '%s is not subclass of %s' % (com_type, Primitive)
            )
        if com_type in self._com_dict:
            raise ComExistError(com_type)

        if len(read_stamp(com_type, COM_DPDC_ATTR)) > 0:
            raise ComDepedencyError(com_type, 'MUST NOT contains dependency')

        self._com_dict[com_type] = []
        self.com_inst[com_type] = com_obj

    def add_com(self, com_type):
        '''
        Put an component under management

        :param class com_type:
        :raise ComExistError:
        '''

        if com_type in self._com_dict:
            raise ComExistError(com_type)

        self._com_dict[com_type] = read_stamp(com_type, COM_DPDC_ATTR)

    def add_coms(self, com_types):
        '''
        Put components under management

        :param list[class] com_types:
        '''

        for com_type in com_types:
            self.add_com(com_type)

    def load(self):
        '''
        Start to create instances of all of components
        '''
        self._expand_com()
        self._mkcom_layer()
        self._mkcom_instance()
        self._loaded = True

    def ref(self, com_type):
        '''
        Return reference to component

        :param class com_type:
        :rtype: object
        '''

        self._must_loaded()
        if com_type not in self.com_inst:
            return None

        return self.com_inst[com_type]

    def sub_ref(self, com_type):
        '''
        Return references to components which extends from com_type

        :param class com_type:
        :rtype: list[object]
        '''

        self._must_loaded()

        coms = []
        for key in self.com_inst:
            if isinstance(self.com_inst[key], com_type):
                coms.append(self.com_inst[key])

        return coms

    def _must_loaded(self):
        if not self._loaded:
            raise InjectorLoadingError()

    def _expand_com(self):
        com_queue = deque(self._com_dict.keys())

        while len(com_queue) > 0:
            com_type = com_queue.popleft()
            req_coms = read_stamp(com_type, COM_DPDC_ATTR)

            if com_type not in self._com_dict:
                self._com_dict[com_type] = req_coms
            for req_com in req_coms:
                if not issubclass(req_com, Component):
                    raise TypeError(
                        '%s is not component in %s' % (req_com, com_type)
                    )
                if req_com not in self._com_dict:
                    com_queue.append(req_com)

    def _mkcom_layer(self):
        com_list = list(self._com_dict.keys())

        while len(com_list) > 0:
            layer = []
            for com_type in com_list:
                in_layer = True
                for req_com in self._com_dict[com_type]:
                    if req_com in com_list:
                        in_layer = False
                        break
                if in_layer is True:
                    layer.append(com_type)
                    com_list.remove(com_type)
            if len(layer) == 0:
                raise CircleComError(com_list)
            self._com_layer.append(layer)

    def _mkcom_instance(self):
        for layer in self._com_layer:
            for com_type in layer:
                if com_type in self.com_inst:
                    continue
                if issubclass(com_type, Primitive):
                    raise PrimError(com_type, 'has not instance')
                arg_types = tuple([t for t in self._com_dict[com_type]])
                args = tuple([self.com_inst[t] for t in arg_types])
                try:
                    self.com_inst[com_type] = com_type(*args)
                except TypeError as e:
                    raise ComCreationError(com_type, args)



HTTP_STATUS_NAMES = {
    100: 'Continue',
    101: 'Switching Protocol',
    200: 'OK',
    201: 'Created',
    202: 'Accepted',
    203: 'Non-Authoritative Information',
    204: 'No Content',
    205: 'Reset Content',
    206: 'Partial Content',
    300: 'Multiple Choices',
    301: 'Moved Permanently',
    302: 'Not Found',
    303: 'See Other',
    304: 'Not Modified',
    305: 'Use Proxy',
    307: 'Temporary Redirect',
    400: 'Bad Request',
    401: 'Unauthorized',
    402: 'Payment Required',
    403: 'Forbidden',
    404: 'Not Found',
    405: 'Method Not Allowed',
    406: 'Not Accepted',
    407: 'Proxy Authentication Required',
    408: 'Request Timeout',
    409: 'Conflict',
    410: 'Gone',
    411: 'Length Required',
    412: 'Precondition Failed',
    413: 'Request Entity Too Large',
    414: 'Request-URI Too Long',
    415: 'Unsupported Media Type',
    416: 'Requested Range Not Satisfiable',
    417: 'Expectation Failed',
    500: 'Internal Server Error',
    501: 'Not Implemented',
    502: 'Bad Gateway',
    503: 'Service Unavailable',
    504: 'Gateway Timeout',
    505: 'HTTP Version Not Supported'
}


def code_to_str(code):
    '''
    Convert HTTP error code to string

    :param int code:
    :raise: HttpStatusError:
    '''

    if code not in HTTP_STATUS_NAMES:
        raise HttpStatusError(code)
    return '%i %s' % (code, HTTP_STATUS_NAMES[code])


class HttpStatusError(Exception):
    def __init__(self, code):
        self.code = code

    def __str__(self):
        return self.code


# only use this error with status in 4xx, 5xx
class HttpError(Exception):
    def __init__(self, status, req, msg=None):
        self.status = status
        self.req = req
        self._msg = 'status=%i, %s %s' % (status, req.method, req.path)
        self.msg = msg

    def __str__(self):
        return self._msg


class Http400Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(400, req, msg)


class Http401Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(401, req, msg)


class Http402Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(402, req, msg)


class Http403Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(403, req, msg)


class Http404Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(404, req, msg)


class Http405Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(405, req, msg)


class Http406Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(406, req, msg)


class Http407Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(407, req, msg)


class Http408Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(408, req, msg)


class Http409Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(409, req, msg)


class Http410Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(410, req, msg)


class Http411Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(411, req, msg)


class Http412Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(412, req, msg)


class Http413Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(413, req, msg)


class Http414Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(414, req, msg)


class Http415Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(415, req, msg)


class Http416Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(416, req, msg)


class Http417Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(417, req, msg)


class Http500Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(500, req, msg)


class Http501Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(501, req, msg)


class Http502Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(502, req, msg)


class Http503Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(503, req, msg)


class Http504Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(504, req, msg)


class Http505Error(HttpError):
    def __init__(self, req, msg=None):
        super().__init__(505, req, msg)


class HttpArgumentError(Exception):
    def __init__(self, arg_str):
        self.arg_str = arg_str

    def __str__(self):
        return self.arg_str



'''
Authentication model
'''

from .std import acc_name, acc_pwd


#:
post_token_pwd = {
    'type': 'object',
    'required': ['grant_type', 'username', 'password'],
    'properties': {
        'grant_type': {'type': 'string', 'pattern': '^password$'},
        'username': acc_name,
        'password': acc_pwd
    }
}

#:
post_token_rtoken = {
    'type': 'object',
    'required': ['grant_type', 'refresh_token'],
    'properties': {
        'grant_type': {'type': 'string', 'pattern': '^refresh_token$'},
        'refresh_token': {'type': 'string'}
    }
}

#:
res_bearer_token = {
    'type': 'object',
    'required': ['token_type', 'expires_in', 'access_token', 'refresh_token'],
    'properties': {
        'token_type': {'type': 'string', 'pattern': 'Bearer'},
        'expires_in': {'type': 'number'},
        'access_token': {'type': 'string'},
        'refresh_token': {'type': 'string'},
    }
}



from abc import ABC

COM_ATTR = '__clink'
COM_DEP = 'req_coms'


class Component(ABC):
    '''
    It doesn't contains anythings. It's use to mark an class become component
    '''

    pass


class Primitive(Component):
    '''
    Primitive components, it isn't create directly by injector. It must be
    add to injector by add_ref(). If other components depend on it, injector
    look up for it's instance, if not founds instance, raise error instead
    of create new one
    '''

    pass



import os
from subprocess import Popen, CalledProcessError
from os import makedirs, remove
from os.path import isdir, isfile, exists, basename, dirname
from shutil import copyfile, copytree, rmtree


def call(args, cwd=None):
    '''
    Perform new processing

    :param list[str] args:
    :param str cwd:
    '''

    exit_code = Popen(args, cwd=cwd).wait()
    if exit_code != 0:
        raise CalledProcessError(exit_code, args)


def cp(src, dest, exist_ignore=True):
    '''
    Copy file or directory

    :param str src:
    :param str dest:
    :param bool exist_ignore:
    '''

    if isdir(src):
        return _cp_dir(src, dest, exist_ignore)
    else:
        return _cp_file(src, dest, exist_ignore)


def rm(path, exist_ignore=True):
    '''
    Remove files

    :param str path:
    :param bool exist_ignore:
    '''

    if (not exists(path)) and exist_ignore:
        return
    if isdir(path):
        rmtree(path)
    else:
        remove(path)


def mkdir(path):
    '''
    Create new directory

    :param str path:
    '''

    if not isdir(path):
        makedirs(path)


def chmod(path, mode):
    '''
    Change mode of file

    :param str path:
    :param int mode:
    '''

    os.chmod(path, mode)


def touch(path, mode):
    '''
    Create new empty file

    :param str path:
    :param int mode:
    '''

    mkdir(dirname(path))
    f = open(path, 'w+')
    f.close()
    os.chmod(path, mode)


def _cp_dir(src, dest, exist_ignore):
    if isdir(dest):
        dest_dir = os.path.join(dest, basename(src))
        if isdir(dest_dir) and exist_ignore:
            return dest_dir
        copytree(src, dest_dir)
        return dest_dir
    else:
        copytree(src, dest)
        return dest


def _cp_file(src, dest, exist_ignore):
    if isfile(dest) and exist_ignore:
        return dest
    if isdir(dest):
        dest_file = os.path.join(dest, basename(src))
        copyfile(src, dest_file)
        return dest_file
    else:
        copyfile(src, dest)
        return dest



from clink import stamp, mapper, Controller
from clink.service import OAuthSv, AccSv, SmtpSv, TemplateSv
from clink.error.http import Http406Error, Http400Error, Http401Error
from clink.util import fs


@stamp(OAuthSv, AccSv, SmtpSv, TemplateSv)
@mapper.path('auth')
class AuthCtl(Controller):
    def __init__(self, oauth_sv, acc_sv, smtp_sv, tpl_sv):
        self._oauth_sv = oauth_sv
        self._acc_sv = acc_sv
        self._smtp_sv = smtp_sv
        self._tpl_sv = tpl_sv

        self._init_tpl_files()

    @mapper.post('token')
    def get_token(self, req, res):
        info = req.body
        if info is None or 'grant_type' not in info:
            raise Http400Error(req, 'Require grant_type')
        grant_type = info['grant_type']

        if grant_type == 'password':
            if 'username' not in info:
                raise Http400Error(req, 'Require username')
            if 'password' not in info:
                raise Http400Error(req, 'Require password')

            acc = self._acc_sv.find_pwd(info['username'], info['password'])
            if acc is None:
                raise Http401Error(req, 'Name or password is invalid')
            res.body = self._oauth_sv.mktoken_pwd(
                info['username'], info['password']
            )
            values = {
                'acc_name': acc['name'],
                'remote_addr': req.remote_addr
            }
            txt_body = self._tpl_sv.build_file(self._LOGIN_TPL, values)

            subject = 'New login'
            self._smtp_sv.send(acc['email'], subject, txt_body)
        elif grant_type == 'refresh_token':
            if 'refresh_token' not in info:
                raise Http400Error(req, 'Require refresh_token')
            res.body = self._oauth_sv.mktoken_rtoken(info['refresh_token'])
        else:
            raise Http406Error(req, 'Not support grant_type=%s' % grant_type)

    def _init_tpl_files(self):
        '''
        Detect template files
        '''

        self._LOGIN_TPL = fs.asset_path('tpl/login.txt')



import inspect
from collections import deque

from clink.com.type import Component


def find(module, com_type=Component):
    '''
    Find class which extends from com_type in module

    :param module module:
    :param class com_type:
    '''

    coms = []
    mod_que = deque([module])

    while len(mod_que) > 0:
        mod = mod_que.popleft()
        for name in dir(mod):
            a = getattr(mod, name)
            if inspect.ismodule(a):
                if a.__name__ != mod.__name__ + '.' + name:
                    continue
                mod_que.append(a)
            elif inspect.isclass(a):
                if a.__module__ != mod.__name__:
                    continue
                if not issubclass(a, com_type):
                    continue
                coms.append(a)
    return coms



_STAMP_KEY = '__clink'
COM_DPDC_ATTR = 'dpdc'


def write_stamp(target, key, value):
    '''
    Write metadata into target

    :param object target:
    :param str key:
    :param object value:
    '''

    if _STAMP_KEY not in dir(target):
        setattr(target, _STAMP_KEY, {})
    stamp = getattr(target, _STAMP_KEY)
    stamp[key] = value


def read_stamp(target, key):
    '''
    Read metadata from target

    :param object target:
    :param str key:
    '''

    if _STAMP_KEY not in dir(target):
        raise TypeError('%s was not stamped' % target)
    stamp = getattr(target, _STAMP_KEY)
    if key not in stamp:
        raise KeyError('%s.%s' % (_STAMP_KEY, key))
    return stamp[key]


def stamp(*args):
    '''
    Mark an class become component with depedencies

    :param tuple args:
    '''

    def decorator_fn(target):
        write_stamp(target, COM_DPDC_ATTR, args)
        return target
    return decorator_fn



import os
from clink.com import read_stamp
from clink.iface import ILv2Handler

from .error import RouteExistError, PathNotFoundError, HandleNotFoundError
from .type import MapNode, NodeAction
from .route import Route
from .mapper import CTL_PATH_ATTR, CTL_METHOD_ATTR


class Router(ILv2Handler):
    '''
    Store and find routes
    '''

    def __init__(self, routes=[]):
        '''
        :param list<Router> routes:
        '''

        self._root_node = MapNode('/')
        self.add_routes(routes)

    def add_ctl(self, ctl):
        '''
        Collect routing information from controller

        :param object ctl:
        '''

        ctl_type = type(ctl)
        ctl_path = read_stamp(ctl, CTL_PATH_ATTR)
        for attr_name in dir(ctl_type):
            ctl_attr = getattr(ctl_type, attr_name)
            try:
                ctl_method = read_stamp(ctl_attr, CTL_METHOD_ATTR)
                abs_path = ctl_path
                if len(ctl_method.path) > 0:
                    abs_path = os.path.join(ctl_path, ctl_method.path)
                route = Route(
                    ctl_method.method, ctl_method.content_type,
                    abs_path, getattr(ctl, attr_name)
                )
                self.add_route(route)
            except KeyError:
                pass
            except TypeError:
                pass

    def add_route(self, route):
        '''
        Put route into map

        :param Route route:
        :raise RouteExistError:
        '''

        # search and add nodes
        node = self._root_node
        if route.path != '/':
            node_names = route.path.split('/')
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    node.child[node_name] = MapNode(node_name)
                node = node.child[node_name]

        # add action to node
        action = NodeAction(route.method, route.content_type, route.handle)
        if self._action_is_exist(node, action):
            raise RouteExistError(route)
        node.actions.append(action)

    def add_routes(self, routes):
        '''
        Put routes into map

        :param list[Route] routes:
        '''

        for route in routes:
            self.add_route(route)

    def handle(self, req):
        '''
        Find handle which match with request

        :param Request req:
        :rtype: function
        :raise PathNotFoundError:
        :raise HandleNotFoundError:
        '''

        node = self._find_node(req.path)
        if node is None or len(node.actions) == 0:
            raise PathNotFoundError(req.path)
        for action in node.actions:
            if action.method != req.method.lower():
                continue
            if action.content_type != req.content_type:
                continue
            return action.handle
        raise HandleNotFoundError(req.method, req.content_type, req.path)

    def _find_node(self, path):
        '''
        notes: algorithm can be improve here
        '''
        node = self._root_node
        if path != '/':
            node_names = path.split('/')
            node_names = list(filter(''.__ne__, node_names))
            for node_name in node_names:
                if node_name not in node.child:
                    return None
                node = node.child[node_name]
        return node

    def _action_is_exist(self, node, action):
        for node_action in node.actions:
            if node_action.method != action.method:
                continue
            if node_action.content_type != action.method:
                continue
            return True
        return False



from inspect import ismodule

from .com import find, Injector
from .iface import IWsgi, IPipeHandler, ILv0Handler, ILv1Handler, ILv3Handler
from .iface import ILv5Handler, ILv6Handler, ILv7Handler, ILv8Handler
from .type import Request, Response, Controller
from .routing import Router
from .handler import RecvHandler, SendHandler
from .handler import ReqJsonHandler, ReqUrlEncodeHandler, ReqLogHandler
from .handler import ResJsonHandler, ResCorsHandler
from .handler import ErrorHttpHandler, ErrorLogHandler, DflowErrorHandler
from .handler import RoutingErrorHandler


class App(IWsgi):
    '''
    Application brings APIs to HTTP
    '''

    _lv0_handler = None
    _lv1_handler = None
    _lv3_handlers = None
    _lv5_handlers = None
    _lv6_handler = None
    _lv7_handlers = None
    _lv8_handler = None
    _lv9_handler = None

    def __init__(self, conf):
        '''
        :param clink.AppConf conf:
        '''

        self.router = Router()
        self.injector = Injector()
        self.injector.add_prim(conf)

        self.injector.add_com(RecvHandler)
        self.injector.add_com(ReqLogHandler)
        self.injector.add_com(ReqJsonHandler)
        self.injector.add_com(ReqUrlEncodeHandler)
        self.injector.add_com(ResJsonHandler)
        self.injector.add_com(ResCorsHandler)
        self.injector.add_com(SendHandler)
        self.injector.add_com(ErrorHttpHandler)
        self.injector.add_com(DflowErrorHandler)
        self.injector.add_com(ErrorLogHandler)
        self.injector.add_com(RoutingErrorHandler)

    def add_handler(self, handler_type):
        '''
        Add handler to application

        :param class handler_type:
        '''

        if not issubclass(handler_type, IPipeHandler):
            raise TypeError('%s is not handler' % handler_type)

        self.injector.add_com(handler_type)

    def add_prim(self, *args):
        for prim_ref in args:
            self.injector.add_prim(prim_ref)

    def add_ctl(self, ctl_type):
        '''
        Add controller to application

        :param class ctl:
        '''

        if not issubclass(ctl_type, Controller):
            raise TypeError('%s is not controller' % ctl_type)

        self.injector.add_com(ctl_type)

    def add_ctls(self, module):
        '''
        Search controllers in module and add them to application
        '''

        if not ismodule(module):
            raise TypeError('%s is not module' % module)

        ctl_types = find(module, Controller)
        self.injector.add_coms(ctl_types)

    def load(self):
        '''
        Creeate instance of all of components, put it to ready state
        '''

        self.injector.load()
        self._init_routes()

        self._lv0_handler = self.injector.sub_ref(ILv0Handler)[0]
        self._lv1_handler = self.injector.sub_ref(ILv1Handler)[0]
        self._lv3_handlers = self.injector.sub_ref(ILv3Handler)
        self._lv5_handlers = self.injector.sub_ref(ILv5Handler)
        self._lv6_handler = self.injector.sub_ref(ILv6Handler)[0]
        self._lv7_handlers = self.injector.sub_ref(ILv7Handler)
        self._lv8_handler = self.injector.sub_ref(ILv8Handler)[0]
        self._lv9_handler = self._lv6_handler

    def __call__(self, wsgi_env, wsgi_send):
        '''
        Implemention of WSGI. That mean you can call instance of App by
        WSGI server to make application is available on network.

        :param dict wsgi_env:
        :param function wsgi_send:
        '''

        # level 0: recv handling
        req = Request()
        res = Response()

        try:
            # level 0 continue: receiving handling
            self._lv0_handler.handle(req, res, wsgi_env)

            # level 1: pre-routing handling
            self._lv1_handler.handle(req, res)

            # level 2: routing, find main handler
            lv4_handle = self.router.handle(req)

            # level 3: pre-main handling
            for handler in self._lv3_handlers:
                handler.handle(req, res)

            # level 4: main handling
            lv4_handle(req, res)

            # level 5: response handling
            for handler in self._lv5_handlers:
                handler.handle(req, res)

            # level 6: send handling
            return self._lv6_handler.handle(req, res, wsgi_send)
        except Exception as e:
            # level 7: error handling
            handled = False
            for handler in self._lv7_handlers:
                if handler.handle(req, res, e):
                    handled = True
            if not handled:
                raise e

            # level 8: error log handling
            self._lv8_handler.handle(req, res, e)

            # level 9: send error response
            return self._lv9_handler.handle(req, res, wsgi_send)

    def _init_routes(self):
        for com_type, com_ref in self.injector.com_inst.items():
            if isinstance(com_ref, Controller):
                self.router.add_ctl(com_ref)



import json

from clink.error.http import code_to_str
from clink.iface import ILv7Handler
from clink.mime.type import MIME_JSON
from clink.com import stamp
from clink.dflow import ExistError, NonExistError, FormatError, ExpiredError


@stamp()
class DflowErrorHandler(ILv7Handler):
    '''
    Catch Data Flow error and make response message correspond with
    error
    '''

    def handle(self, req, res, e):
        if isinstance(e, ExistError):
            res.status = 409
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 409,
                'status_name': code_to_str(409),
                'message': e.indexes
            }).encode('utf-8')
            return True
        elif isinstance(e, NonExistError):
            res. status = 404
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 404,
                'status_name': code_to_str(404),
                'message': e.indexes
            }).encode('utf-8')
            return True
        elif isinstance(e, FormatError):
            res.status = 400
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 400,
                'status_name': code_to_str(400),
                'message': {
                    'name': e.name, 'value': str(e.value), 'schema': e.schema
                }
            }).encode('utf-8')
            return True
        elif isinstance(e, ExpiredError):
            res.status = 403
            res.header = {}
            res.content_type = MIME_JSON
            res.body = json.dumps({
                'status': 403,
                'status_name': code_to_str(403),
                'message': e.indexes
            }).encode('utf-8')
            return True
        else:
            return False



class CtlMethod():
    '''
    Specify controller method
    '''

    def __init__(self, method, path, content_type):
        '''
        :param str method:
        :param str path:
        :param str content_type:
        '''

        self.method = method
        self.path = path
        self.content_type = content_type


class NodeAction():
    '''
    Specify pair of method, content type and handle in a map node
    '''

    def __init__(self, method, content_type, handle):
        '''
        :param str method:
        :param str content_type:
        :param function handle:
        '''

        self.method = method
        self.content_type = content_type
        self.handle = handle


class MapNode():
    '''
    Specify a map node with name, children and actions
    '''

    def __init__(self, name, child={}, actions=[]):
        '''
        :param str name:
        :param list[MapNode] child:
        :param list[NodeAction] actions:
        '''

        self.name = name
        self.child = {}
        self.actions = []



from abc import ABC, abstractmethod


class IWsgi(ABC):
    @abstractmethod
    def __call__(self, wsgi_env, wsgi_send):
        '''
        WSGI inteface

        :param dict wsgi_env:
        :param function wsgi_send:
        '''

        pass



class ConfirmCodeSpec():
    '''
    Specify confirm code and expired date
    '''

    def __init__(self, code, exp_date):
        '''
        :param str code:
        :param datetime exp_date:
        '''

        self.code = code
        self.exp_date = exp_date
